from dataclasses import dataclass

from bq_schema.bigquery_table import BigqueryTable


@dataclass
class FirstTableSchema:
    a_column: int


class FirstTable(BigqueryTable):
    name = "first_table"
    schema = FirstTableSchema



from google.cloud.bigquery import SchemaField

from bq_schema.bigquery_table import BigqueryTable


class SecondTable(BigqueryTable):
    name = "second_table"
    schema = [SchemaField("b_column", "FLOAT")]



"""
Convert a python dataclass into a BigQuery schema definition.
"""
from dataclasses import Field, fields, is_dataclass
from typing import Any, List, Optional, Type, Union

from google.cloud.bigquery import SchemaField

from bq_schema.types.type_parser import (
    parse_inner_type_of_list,
    parse_inner_type_of_optional,
)

from .types import BigQueryFieldModes, BigQueryTypes, TypeMapping

_BASIC_TYPES_TO_NAME = {
    primitive_type: bq_type for bq_type, primitive_type in TypeMapping.items()
}
_NoneType = type(None)


def dataclass_to_schema(dataclass: Type) -> List[SchemaField]:
    """
    Transfrom a dataclass into a list of SchemaField.
    """
    if not is_dataclass(dataclass):
        raise TypeError("Not a dataclass.")

    return [_field_to_schema(field) for field in fields(dataclass)]


def _field_to_schema(field: Field) -> SchemaField:
    field_type = _BASIC_TYPES_TO_NAME.get(field.type)
    if field_type:
        return SchemaField(
            name=field.name,
            field_type=field_type,
            description=_parse_field_description(field),
            mode=BigQueryFieldModes.REQUIRED,
        )

    if is_dataclass(field.type):
        return SchemaField(
            name=field.name,
            field_type=BigQueryTypes.STRUCT,
            mode=BigQueryFieldModes.REQUIRED,
            description=_parse_field_description(field),
            fields=_parse_fields(field.type),
        )

    # typing.Optional is the same as typing.Union[SomeType, NoneType]
    if field.type.__origin__ is Union:
        return _parse_optional(field)

    if field.type.__origin__ is list:
        return _parse_list(field)

    raise TypeError(f"Unsupported type: {field.type}.")


def _parse_list(field: Field) -> SchemaField:
    field_type = parse_inner_type_of_list(field.type)
    return SchemaField(
        name=field.name,
        field_type=_python_type_to_big_query_type(field_type),
        mode=BigQueryFieldModes.REPEATED,
        description=_parse_field_description(field),
        fields=_parse_fields(field_type),
    )


def _parse_optional(field: Field) -> SchemaField:
    field_type = parse_inner_type_of_optional(field.type)
    return SchemaField(
        name=field.name,
        field_type=_python_type_to_big_query_type(field_type),
        mode=BigQueryFieldModes.NULLABLE,
        description=_parse_field_description(field),
        fields=_parse_fields(field_type),
    )


def _parse_fields(field_type: Type) -> List[SchemaField]:
    if is_dataclass(field_type):
        return dataclass_to_schema(field_type)

    return []


def _python_type_to_big_query_type(field_type: Any) -> BigQueryTypes:
    if is_dataclass(field_type):
        return BigQueryTypes.STRUCT

    bq_type = _BASIC_TYPES_TO_NAME.get(field_type)
    if bq_type:
        return bq_type

    raise TypeError(f"Unsupported type: {field_type}")


def _parse_field_description(field: Field) -> Optional[str]:
    if "description" in field.metadata:
        return field.metadata["description"]
    return None



from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from typing import List, Optional

from google.cloud.bigquery.table import Row

from bq_schema.row_transformer import RowTransformer
from bq_schema.types.type_mapping import Geography, Timestamp


@dataclass
class NestedAgain:
    int_field: int


@dataclass
class NestedSchema:
    int_field: int
    nested_again: NestedAgain
    nested_again_optional_set: Optional[NestedAgain]
    nested_again_optional_unset: Optional[NestedAgain]
    nested_again_repeated: List[NestedAgain]


# pylint: disable=too-many-instance-attributes
@dataclass
class Schema:
    string_field: str
    bytes_field: bytes
    int_field: int
    float_field: float
    numeric_field: Decimal
    bool_field: bool
    timestamp_field: Timestamp
    date_field: date
    time_field: time
    optional_int_unset: Optional[int]
    optional_int_set: Optional[int]
    optional_nested_field: Optional[NestedSchema]
    nested_field: NestedSchema
    nested_repeated_field: List[NestedSchema]
    datetime_field: datetime
    geography_field: Geography


def test_transform_dataclass_instance_to_row():

    instance = Schema(
        string_field="string",
        bytes_field=b"string",
        int_field=1,
        float_field=1.0,
        numeric_field=Decimal("1.0"),
        bool_field=True,
        timestamp_field=Timestamp(datetime(2020, 1, 1)),
        date_field=date(2020, 1, 1),
        time_field=time(1, 1, 1),
        optional_int_unset=None,
        optional_int_set=10,
        optional_nested_field=None,
        datetime_field=datetime(2020, 2, 2),
        geography_field=Geography("POINT (10 10)"),
        nested_field=NestedSchema(
            int_field=2,
            nested_again=NestedAgain(int_field=3),
            nested_again_optional_set=NestedAgain(int_field=3),
            nested_again_optional_unset=None,
            nested_again_repeated=[NestedAgain(int_field=4)],
        ),
        nested_repeated_field=[
            NestedSchema(
                int_field=2,
                nested_again=NestedAgain(int_field=3),
                nested_again_optional_set=NestedAgain(int_field=3),
                nested_again_optional_unset=None,
                nested_again_repeated=[NestedAgain(int_field=4)],
            )
        ],
    )
    row_transformer = RowTransformer(Schema)
    expected = {
        "bool_field": True,
        "bytes_field": b"string",
        "date_field": date(2020, 1, 1),
        "datetime_field": datetime(2020, 2, 2, 0, 0),
        "geography_field": "POINT (10 10)",
        "float_field": 1.0,
        "int_field": 1,
        "nested_field": {
            "int_field": 2,
            "nested_again": {"int_field": 3},
            "nested_again_optional_set": {"int_field": 3},
            "nested_again_optional_unset": None,
            "nested_again_repeated": [{"int_field": 4}],
        },
        "nested_repeated_field": [
            {
                "int_field": 2,
                "nested_again": {"int_field": 3},
                "nested_again_optional_set": {"int_field": 3},
                "nested_again_optional_unset": None,
                "nested_again_repeated": [{"int_field": 4}],
            }
        ],
        "numeric_field": Decimal("1.0"),
        "string_field": "string",
        "time_field": time(1, 1, 1),
        "optional_int_set": 10,
        "optional_int_unset": None,
        "optional_nested_field": None,
        "timestamp_field": Timestamp(datetime(2020, 1, 1, 0, 0)),
    }
    assert row_transformer.dataclass_instance_to_bq_row(instance) == expected


def test_transfrom_row_into_dataclass():
    row_as_dict = {
        "bool_field": True,
        "bytes_field": b"string",
        "date_field": date(2020, 1, 1),
        "datetime_field": datetime(2020, 2, 2, 0, 0),
        "geography_field": "POINT(10 10)",
        "float_field": 1.0,
        "int_field": 1,
        "nested_field": {
            "int_field": 2,
            "nested_again": {"int_field": 3},
            "nested_again_optional_set": {"int_field": 3},
            "nested_again_repeated": [{"int_field": 4}],
        },
        "nested_repeated_field": [
            {
                "int_field": 2,
                "nested_again": {"int_field": 3},
                "nested_again_optional_set": {"int_field": 3},
                "nested_again_repeated": [{"int_field": 4}],
            }
        ],
        "numeric_field": Decimal("1.0"),
        "string_field": "string",
        "time_field": time(1, 1, 1),
        "optional_int_set": 10,
        "timestamp_field": Timestamp(datetime(2020, 1, 1, 0, 0)),
    }
    row_transformer = RowTransformer(Schema)
    row = dict_to_row(row_as_dict)

    expected = Schema(
        string_field="string",
        bytes_field=b"string",
        int_field=1,
        float_field=1.0,
        numeric_field=Decimal("1.0"),
        bool_field=True,
        timestamp_field=Timestamp(datetime(2020, 1, 1)),
        date_field=date(2020, 1, 1),
        time_field=time(1, 1, 1),
        optional_int_unset=None,
        optional_int_set=10,
        optional_nested_field=None,
        datetime_field=datetime(2020, 2, 2),
        geography_field=Geography("POINT(10 10)"),
        nested_field=NestedSchema(
            int_field=2,
            nested_again=NestedAgain(int_field=3),
            nested_again_optional_set=NestedAgain(int_field=3),
            nested_again_optional_unset=None,
            nested_again_repeated=[NestedAgain(int_field=4)],
        ),
        nested_repeated_field=[
            NestedSchema(
                int_field=2,
                nested_again=NestedAgain(int_field=3),
                nested_again_optional_set=NestedAgain(int_field=3),
                nested_again_optional_unset=None,
                nested_again_repeated=[NestedAgain(int_field=4)],
            )
        ],
    )
    assert expected == row_transformer.bq_row_to_dataclass_instance(row)


def dict_to_row(row_as_dict: dict) -> Row:
    values = []
    field_to_index = {}
    for i, (key, value) in enumerate(row_as_dict.items()):
        values.append(value)
        field_to_index[key] = i

    return Row(values=values, field_to_index=field_to_index)



from dataclasses import asdict
from typing import Generic, Type, TypeVar

from dacite import Config, from_dict
from google.cloud.bigquery.table import Row

T = TypeVar("T")  # pylint: disable=invalid-name


class RowTransformer(Generic[T]):
    """
    Serialized / deserialize rows.
    """

    def __init__(self, schema: Type[T]):
        self._schema: Type[T] = schema

    def bq_row_to_dataclass_instance(self, bq_row: Row) -> T:
        """
        Create a dataclass instance from a row returned by the bq library.
        """
        return from_dict(self._schema, bq_row, config=Config(check_types=False))

    @staticmethod
    def dataclass_instance_to_bq_row(instance: T) -> dict:
        """
        Convert a dataclass instance into a dictionary, which can be inserted into bq.
        """
        return asdict(instance)



# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass, field
from datetime import date, datetime, time
from decimal import Decimal
from typing import List, Optional

from google.cloud.bigquery import SchemaField

from bq_schema.dataclass_converter import dataclass_to_schema
from bq_schema.types.type_mapping import Geography, Timestamp


def test_types():
    @dataclass
    class NestedSchema:
        int_field: int = field(metadata={"description": "This is an INT field."})

    @dataclass
    class Schema:
        string_field: str = field(metadata={"description": "This is a STRING field."})
        bytes_field: bytes
        int_field: int
        float_field: float
        numeric_field: Decimal
        bool_field: bool
        timestamp_field: Timestamp
        date_field: date
        time_field: time
        datetime_field: datetime
        geography_field: Geography
        nested_field: NestedSchema = field(
            metadata={"description": "This is a STRUCT field."}
        )

    assert dataclass_to_schema(Schema) == [
        SchemaField(
            "string_field", "STRING", "REQUIRED", "This is a STRING field.", ()
        ),
        SchemaField("bytes_field", "BYTES", "REQUIRED", None, ()),
        SchemaField("int_field", "INT64", "REQUIRED", None, ()),
        SchemaField("float_field", "FLOAT64", "REQUIRED", None, ()),
        SchemaField("numeric_field", "NUMERIC", "REQUIRED", None, ()),
        SchemaField("bool_field", "BOOL", "REQUIRED", None, ()),
        SchemaField("timestamp_field", "TIMESTAMP", "REQUIRED", None, ()),
        SchemaField("date_field", "DATE", "REQUIRED", None, ()),
        SchemaField("time_field", "TIME", "REQUIRED", None, ()),
        SchemaField("datetime_field", "DATETIME", "REQUIRED", None, ()),
        SchemaField("geography_field", "GEOGRAPHY", "REQUIRED", None, ()),
        SchemaField(
            "nested_field",
            "STRUCT",
            "REQUIRED",
            "This is a STRUCT field.",
            (
                SchemaField(
                    "int_field", "INT64", "REQUIRED", "This is an INT field.", ()
                ),
            ),
        ),
    ]


def test_optional_types():
    @dataclass
    class NestedSchema:
        int_field: Optional[int] = field(
            metadata={"description": "This is an INT field."}
        )

    @dataclass
    class Schema:
        string_field: Optional[str] = field(
            metadata={"description": "This is a STRING field."}
        )
        bytes_field: Optional[bytes]
        int_field: Optional[int]
        float_field: Optional[float]
        numeric_field: Optional[Decimal]
        bool_field: Optional[bool]
        timestamp_field: Optional[Timestamp]
        date_field: Optional[date]
        time_field: Optional[time]
        datetime_field: Optional[datetime]
        geography_field: Optional[Geography]
        nested_field: Optional[NestedSchema] = field(
            metadata={"description": "This is a STRUCT field."}
        )

    assert dataclass_to_schema(Schema) == [
        SchemaField(
            "string_field", "STRING", "NULLABLE", "This is a STRING field.", ()
        ),
        SchemaField("bytes_field", "BYTES", "NULLABLE", None, ()),
        SchemaField("int_field", "INT64", "NULLABLE", None, ()),
        SchemaField("float_field", "FLOAT64", "NULLABLE", None, ()),
        SchemaField("numeric_field", "NUMERIC", "NULLABLE", None, ()),
        SchemaField("bool_field", "BOOL", "NULLABLE", None, ()),
        SchemaField("timestamp_field", "TIMESTAMP", "NULLABLE", None, ()),
        SchemaField("date_field", "DATE", "NULLABLE", None, ()),
        SchemaField("time_field", "TIME", "NULLABLE", None, ()),
        SchemaField("datetime_field", "DATETIME", "NULLABLE", None, ()),
        SchemaField("geography_field", "GEOGRAPHY", "NULLABLE", None, ()),
        SchemaField(
            "nested_field",
            "STRUCT",
            "NULLABLE",
            "This is a STRUCT field.",
            (
                SchemaField(
                    "int_field", "INT64", "NULLABLE", "This is an INT field.", ()
                ),
            ),
        ),
    ]


def test_repeated_types():
    @dataclass
    class NestedSchema:
        int_field: List[int] = field(metadata={"description": "This is an INT field."})

    @dataclass
    class Schema:
        string_field: List[str] = field(
            metadata={"description": "This is a STRING field."}
        )
        bytes_field: List[bytes]
        int_field: List[int]
        float_field: List[float]
        numeric_field: List[Decimal]
        bool_field: List[bool]
        timestamp_field: List[Timestamp]
        date_field: List[date]
        time_field: List[time]
        datetime_field: List[datetime]
        geography_field: List[Geography]
        nested_field: List[NestedSchema] = field(
            metadata={"description": "This is a STRUCT field."}
        )

    assert dataclass_to_schema(Schema) == [
        SchemaField(
            "string_field", "STRING", "REPEATED", "This is a STRING field.", ()
        ),
        SchemaField("bytes_field", "BYTES", "REPEATED", None, ()),
        SchemaField("int_field", "INT64", "REPEATED", None, ()),
        SchemaField("float_field", "FLOAT64", "REPEATED", None, ()),
        SchemaField("numeric_field", "NUMERIC", "REPEATED", None, ()),
        SchemaField("bool_field", "BOOL", "REPEATED", None, ()),
        SchemaField("timestamp_field", "TIMESTAMP", "REPEATED", None, ()),
        SchemaField("date_field", "DATE", "REPEATED", None, ()),
        SchemaField("time_field", "TIME", "REPEATED", None, ()),
        SchemaField("datetime_field", "DATETIME", "REPEATED", None, ()),
        SchemaField("geography_field", "GEOGRAPHY", "REPEATED", None, ()),
        SchemaField(
            "nested_field",
            "STRUCT",
            "REPEATED",
            "This is a STRUCT field.",
            (
                SchemaField(
                    "int_field", "INT64", "REPEATED", "This is an INT field.", ()
                ),
            ),
        ),
    ]



import inspect
import os
from importlib import util as importlib_util
from typing import Iterator, Set

from bq_schema.bigquery_table import BigqueryTable


def find_tables(module_path: str) -> Set[BigqueryTable]:
    tables = {}
    for table in _tables_iterator(module_path):
        tables[(table.project, table.dataset, table.name)] = table

    return set(tables.values())


def _tables_iterator(module_path: str) -> Iterator[BigqueryTable]:
    """
    Recursively find all defined bigquery tables.
    """
    for (dir_path, _, file_names) in os.walk(module_path):
        for file_name in file_names:
            if file_name.endswith(".py"):
                spec = importlib_util.spec_from_file_location(
                    "not_important", os.path.join(dir_path, file_name)
                )
                assert spec is not None
                mod = importlib_util.module_from_spec(spec)

                spec.loader.exec_module(mod)  # type: ignore

                for _, obj in inspect.getmembers(mod, inspect.isclass):
                    if issubclass(obj, BigqueryTable) and not obj == BigqueryTable:
                        instance = obj()
                        yield instance



import argparse
from argparse import Namespace

from bq_schema.cli.bigquery_connection import create_connection
from bq_schema.schema_converter import schema_to_dataclass


def parse_args() -> Namespace:
    parser = argparse.ArgumentParser(
        description="Convert a bigquery table schema into a dataclass."
    )
    parser.add_argument("--project", required=True, help="Target bigquery project.")
    parser.add_argument("--dataset", required=True, help="Target bigquery dataset")
    parser.add_argument(
        "--table-name", required=True, help="Target bigquery table_name"
    )

    return parser.parse_args()


def main(project: str, dataset: str, table_name: str) -> None:
    client = create_connection()
    table = client.get_table(f"{project}.{dataset}.{table_name}")
    print("from dataclasses import dataclass")
    print(schema_to_dataclass(table_name, table.schema))


def cli() -> None:
    args = parse_args()
    main(args.project, args.dataset, args.table_name)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['bq_schema', 'bq_schema.migration', 'bq_schema.types']

package_data = \
{'': ['*'], 'bq_schema': ['cli/*']}

install_requires = \
['google-cloud-bigquery==1.24.0', 'dacite==1.5.1']

extras_require = \
{'develop': ['invoke==1.4.1'],
 'test': ['black==20.8b1',
          'isort==5.6.4',
          'mypy==0.782',
          'pylint==2.6.0',
          'pytest==6.0.2',
          'pytest-cov==2.10.1']}

entry_points = \
{'console_scripts': ['convert-table = bq_schema.cli.convert_table:cli',
                     'migrate-tables = bq_schema.cli.migrate_tables:cli']}

setup(name='bq-schema',
      version='0.5.5',
      description='Define your BigQuery tables as dataclasses.',
      author='Limehome GmbH',
      author_email='raphael.brand@limehome.de',
      url='https://github.com/limehome/bq-schema',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



from typing import Any, Type

_NoneType = type(None)


def parse_inner_type_of_list(list_type: Any) -> Type:
    return list_type.__args__[0]


def parse_inner_type_of_optional(optional_type: Any) -> Type:

    args = optional_type.__args__
    if not (len(args) == 2 and any(arg is _NoneType for arg in args)):
        raise TypeError(f"Unsupported type: {optional_type}.")

    return next(arg for arg in args if arg is not _NoneType)



from enum import Enum


class BigQueryTypes(str, Enum):
    STRING = "STRING"
    BYTES = "BYTES"
    INTEGER = "INT64"
    INT64 = "INT64"
    FLOAT64 = "FLOAT64"
    FLOAT = "FLOAT64"
    NUMERIC = "NUMERIC"
    BOOL = "BOOL"
    BOOLEAN = "BOOL"
    STRUCT = "STRUCT"
    RECORD = "STRUCT"
    TIMESTAMP = "TIMESTAMP"
    DATE = "DATE"
    TIME = "TIME"
    DATETIME = "DATETIME"
    GEOGRAPHY = "GEOGRAPHY"



from typing import Dict, Iterator, List, Optional, Union

from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import Client as BigQueryClient
from google.cloud.bigquery.schema import _STRUCT_TYPES, LEGACY_TO_STANDARD_TYPES
from google.cloud.bigquery.table import Table
from google.cloud.bigquery_v2 import types

from bq_schema.migration.models import ExistingTable, MissingTable
from bq_schema.migration.table_finder import find_tables

_SchemaDiffs = Dict[str, Union[MissingTable, ExistingTable]]

# pylint: disable=no-member
def check_schemas(
    local_schema: List[bigquery.SchemaField],
    remote_schema: List[bigquery.SchemaField],
    is_nested: Optional[bool] = False,
) -> Iterator[str]:
    remote_columns = {column.name: column for column in remote_schema}
    local_columns = {column.name: column for column in local_schema}

    for column in remote_schema:

        if column.name not in local_columns:
            yield f"Nested: Removed column {column}" if is_nested else f"Removed {column}"
            continue

    for column in local_schema:

        if column.name not in remote_columns:
            yield f"Nested: New column {column}" if is_nested else f"New column {column}"
            continue
        modes_not_equal = column.mode != remote_columns[column.name].mode
        column_field_type = LEGACY_TO_STANDARD_TYPES.get(
            column.field_type, types.StandardSqlDataType.TypeKind.TYPE_KIND_UNSPECIFIED
        )
        remote_column_field_type = LEGACY_TO_STANDARD_TYPES.get(
            remote_columns[column.name].field_type,
            types.StandardSqlDataType.TypeKind.TYPE_KIND_UNSPECIFIED,
        )
        if types.StandardSqlDataType.TypeKind.TYPE_KIND_UNSPECIFIED in (
            column_field_type,
            remote_column_field_type,
        ):
            check_schema_msg = (
                f"Unspecified field type in {column} or {remote_columns[column.name]}"
            )
            yield check_schema_msg
            continue
        field_types_not_equal = column_field_type != remote_column_field_type
        if modes_not_equal or field_types_not_equal:
            check_schema_msg = (
                f"Nested: There is difference between {column} and {remote_columns[column.name]}"
                if is_nested
                else f"There is difference between {column} and {remote_columns[column.name]}"
            )
            yield check_schema_msg
            continue

        if column.field_type in _STRUCT_TYPES:
            yield from check_schemas(
                column.fields, remote_columns[column.name].fields, True
            )


def find_schema_differences(
    module_path: str,
    bigquery_client: BigQueryClient,
    global_project: Optional[str],
    global_dataset: Optional[str],
) -> _SchemaDiffs:
    schema_diffs: _SchemaDiffs = {}
    for local_table in find_tables(module_path):
        project = global_project or local_table.project
        assert project, "Project has not been set."
        dataset = global_dataset or local_table.dataset
        assert dataset, "Dataset has not been set."

        table_identifier = f"{project}.{dataset}.{local_table.full_table_name()}"

        try:
            remote_table = bigquery_client.get_table(table_identifier)
            if list(
                check_schemas(local_table.get_schema_fields(), remote_table.schema)
            ):
                schema_diffs[table_identifier] = ExistingTable(
                    local_table=local_table,
                    remote_table=remote_table,
                    schema_diffs=list(
                        check_schemas(
                            local_table.get_schema_fields(), remote_table.schema
                        )
                    ),
                )
        except NotFound:
            schema_diffs[table_identifier] = MissingTable(local_table=local_table)

    return schema_diffs


def print_format_schema_differences(schema_diffs: _SchemaDiffs) -> Dict[str, str]:
    formated_prints = {}
    for table_identifier, difference in schema_diffs.items():
        if isinstance(difference, MissingTable):
            formated_prints[table_identifier] = "Table does not exist in bq"
        elif isinstance(difference, ExistingTable):
            formated_prints[
                table_identifier
            ] = f"Schema differences: {difference.schema_diffs}"

    return formated_prints


def confirm_apply_schema_differences() -> bool:
    valid = {"yes": True, "y": True, "no": False, "n": False}
    while True:
        confirm_apply = (
            input("Do you really want to apply the changes? [Y/N]: ").lower().strip()
        )
        if confirm_apply in valid:
            return valid[confirm_apply]

        print("Possible choices: [Y, y, Yes, yes, N, n, No, no]")


def apply_schema_differences(
    schema_diffs: _SchemaDiffs,
    bigquery_client: BigQueryClient,
) -> None:
    print("Applying changes...")
    for table_identifier, difference in schema_diffs.items():
        if isinstance(difference, MissingTable):
            print("Creating table...")
            table = Table(
                table_identifier,
                schema=difference.local_table.get_schema_fields(),
            )
            if difference.local_table.time_partitioning:
                table.time_partitioning = difference.local_table.time_partitioning
            remote_table = bigquery_client.create_table(table)
            print(remote_table)
        elif isinstance(difference, ExistingTable):
            difference.remote_table.schema = difference.local_table.get_schema_fields()
            print(bigquery_client.update_table(difference.remote_table, ["schema"]))



"""
Convert a bigquery schema into the source code for a data class.
"""
from typing import List

from google.cloud.bigquery import SchemaField

from .types import BigQueryFieldModes, BigQueryTypes, TypeMapping

_TEMPLATE_DATACLASS = """
@dataclass
class {schema_name}:
{fields}
"""
_TEMPLATE_FIELD = "    {field_name}: {field_type}"
_TEMPLATE_FIELD_WITH_DESCRIPTION = (
    "    {field_name}: {field_type} ="
    ' field(metadata={{"description": "{field_description}"}})'
)

_STRUCT_TYPES = {"RECORD", "STRUCT"}


def schema_to_dataclass(schema_name: str, schema: List[SchemaField]) -> str:
    """
    Convert a list of schema fields, our schema, into a dataclass.
    """

    struct_fields = [s for s in schema if s.field_type in _STRUCT_TYPES]
    nested_dataclasses = "".join(
        schema_to_dataclass(s.name, s.fields) for s in struct_fields
    )
    dataclass_fields = [_create_field(field) for field in schema]
    python_dataclass = _TEMPLATE_DATACLASS.format(
        schema_name=_generate_dataclass_name(schema_name),
        fields="\n".join(dataclass_fields),
    ).strip()

    return (
        f"{nested_dataclasses}\n{python_dataclass}"
        if nested_dataclasses
        else f"{python_dataclass}\n"
    )


def _create_field(field: SchemaField) -> str:
    bigquery_type = BigQueryTypes[field.field_type]
    if bigquery_type == BigQueryTypes.STRUCT:
        dataclass_type = _generate_dataclass_name(field.name)
    else:
        dataclass_type = TypeMapping[bigquery_type].__name__

    field_mode = BigQueryFieldModes[field.mode]

    if field_mode == BigQueryFieldModes.REPEATED:
        dataclass_type = f"List[{dataclass_type}]"
    if field_mode == BigQueryFieldModes.NULLABLE:
        dataclass_type = f"Optional[{dataclass_type}]"

    field_string: dict = {
        "template": _TEMPLATE_FIELD,
        "args": {
            "field_name": field.name,
            "field_type": dataclass_type,
        },
    }

    if field.description:
        field_string["template"] = _TEMPLATE_FIELD_WITH_DESCRIPTION
        field_string["args"]["field_description"] = field.description

    return field_string["template"].format(**field_string["args"])


def _generate_dataclass_name(schema_name: str) -> str:
    return (
        "".join(word.capitalize() for word in schema_name.split("_"))
        if "_" in schema_name
        else schema_name[0].upper() + schema_name[1:]
    )



from dataclasses import dataclass
from typing import List

from google.cloud.bigquery.table import Table

from bq_schema.bigquery_table import BigqueryTable


@dataclass
class MissingTable:
    local_table: BigqueryTable


@dataclass
class ExistingTable:
    local_table: BigqueryTable
    remote_table: Table
    schema_diffs: List[str]
